import os
import inspect
import types
import re
import fnmatch
import functools
import warnings
import random

string_types = (type(b""), type(u""))


def invert_depends_dictionary(depends_dictionary):
    inverse_dictionary = {}
    for target in depends_dictionary:
        inverse_dictionary[target] = set()

    for target in depends_dictionary:
        deps_of_target = depends_dictionary[target]
        for dep in deps_of_target:
            inverse_dictionary[dep].add(target)
    return inverse_dictionary


def random_string(length):
    return "".join(random.choice("0123456789abcdefghijklmnopqrstuvwxyz") for i in range(length))


def deprecated(reason):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                warnings.simplefilter("default", DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))


def error(str):
    print(red("LicantError: ") + str)
    exit(-1)


def error_exception(str, ex):
    import traceback

    print(red("LicantException: ") + str)
    print(ex.__class__)
    traceback.print_exc()
    exit(-1)


def cutinvoke(func, *args, **kwargs):
    if isinstance(func, types.FunctionType):
        ins = inspect.getargspec(func)
        nargs = len(ins.args)
        return func(*args[:nargs])
    else:
        return func(*args)


class quite:
    pass


class queue:
    class DontHaveArg:
        pass

    def __init__(self):
        self.lst = []
        self.rdr = 0

    def put(self, obj):
        self.lst.append(obj)

    def get(self):
        if len(self.lst) == 0:
            raise Exception("DontHaveArg")

        ret = self.lst[self.rdr]

        self.rdr += 1
        if self.rdr >= len(self.lst):
            self.__init__()

        return ret

    def empty(self):
        return len(self.lst) == 0

    def __str__(self):
        return str(self.lst)


def textblock(str):
    return chr(27) + str + chr(27) + "[0m"


def black(str):
    return textblock("[30;1m" + str)


def red(str):
    return textblock("[31;1m" + str)


def green(str):
    return textblock("[32;1m" + str)


def yellow(str):
    return textblock("[33;1m" + str)


def purple(str):
    return textblock("[35;1m" + str)


def cyan(str):
    return textblock("[36;1m" + str)


def white(str):
    return textblock("[37;1m" + str)


def always_true(*args, **kwargs):
    return True


def always_false(*args, **kwargs):
    return False


def do_nothing(*args, **kwargs):
    pass


def changeext(path, newext):
    return os.path.splitext(path)[0] + "." + newext


def flag_prefix(pref, lst):
    if lst and lst != []:
        return " ".join(map(lambda x: pref + x, lst))
    else:
        return ""


def find_recursive(root, pattern, hide, debug):
    result = []

    if hide is None:
        for d, dirs, files in os.walk(root):
            for f in files:
                if pattern in f:
                    path = os.path.join(d, f)
                    if debug:
                        print(path)
                    result.append(path)

    else:
        for d, dirs, files in os.walk(root):
            if hide not in d:
                for f in files:
                    if pattern in f:
                        path = os.path.join(d, f)
                        if debug:
                            print(path)
                        result.append(path)

    return result


cxx_read_depends_pattern = re.compile(r"[\w./-]+")


def cxx_read_depends(path):
    if not os.path.exists(path):
        return None
    else:
        f = open(path)
        text = f.read()

        if len(text) == 0:
            return None

        lst = cxx_read_depends_pattern.findall(text)
        return lst[2:]


def get_actions(target):
    clsactions = list(target.__actions__)
    # objactions = [k for k in target.__dict__.keys() if callable(target.__dict__[k])]
    return sorted(clsactions)

    # return target.__dict__


def recursive_glob(treeroot, pattern):
    results = []
    for base, dirs, files in os.walk(treeroot):
        goodfiles = fnmatch.filter(files, pattern)
        results.extend(os.path.join(base, f) for f in goodfiles)
    return results


def canonical_path(path):
    return os.path.abspath(os.path.normpath(path))



import os
import sys
import licant.make
import licant.util
import subprocess
from licant.core import default_core

error_in_install_library = False
termux_dir = "/data/data/com.termux/files"

is_termux = "ANDROID_ROOT" in os.environ
is_windows = sys.platform == 'win32'


def find_application_path():
    global error_in_install_library

    envout = os.environ["PATH"]
    path_list = envout.split(":")

    if "PREFIX" in os.environ:
        return os.path.join(os.environ["PREFIX"], "bin")

    if "/usr/local/bin" in path_list:
        path = "/usr/local/bin"

    else:
        for p in path_list:
            if "/usr/bin" in p:
                path = p
                break
        for p in path_list:
            if "/usr/local/bin" in p:
                path = p
                break
        for p in path_list:
            if "\\usr\\local\\bin" in p:
                path = p
                break
        else:
            print("Warning: Install path not found")
            error_in_install_library = True

    return path


def find_headers_path():
    global error_in_install_library

    if "PREFIX" in os.environ:
        return os.path.join(os.environ["PREFIX"], "include")

    # if is_termux:
    #    return os.path.join(termux_dir, "usr/include")

    if is_windows:
        return "/usr/local/include"

    return "/usr/local/include"


def find_libraries_path():
    global error_in_install_library

    if "PREFIX" in os.environ:
        return os.path.join(os.environ["PREFIX"], "lib")

    if is_windows:
        return "/usr/lib"

    return "/usr/lib"


path = find_application_path()
headers_path = find_headers_path()
libraries_path = find_libraries_path()


def install_application(src, dst, tgt=None):
    if error_in_install_library:
        return None

    apptgt = os.path.join(path, dst)
    if tgt is None:
        tgt = apptgt
    licant.make.copy(tgt=apptgt, src=src)
    return licant.fileset(tgt=tgt, targets=[apptgt])


def install_headers(tgtdir, srcdir, patterns=("*.h", "*.hxx"), adddeps=[]):
    srcdir = os.path.abspath(srcdir)
    lsts = [licant.util.recursive_glob(
        os.path.abspath(srcdir), p) for p in patterns]
    # print(lsts)

    headers = []
    for lst in lsts:
        headers.extend(lst)

    for h in headers:
        licant.source(h)

    targets = [licant.copy(src=h, tgt=os.path.join(
        headers_path, tgtdir, os.path.relpath(h, srcdir)), adddeps=adddeps) for h in headers]
    full_target = licant.fileset(tgt="headers://"+srcdir, targets=targets)

    return full_target, targets


def install_shared_library(src, newname=None):
    if error_in_install_library:
        return None

    if newname is None:
        newname = os.path.basename(src)

    tgt = os.path.join(libraries_path, newname)
    licant.make.copy(tgt=tgt, src=src)

    return tgt


def install_library(tgt, hroot, headers, headers_patterns=("*.h", "*.hxx", "*.hpp"), uninstall=None, libtgt=None):
    if error_in_install_library:
        return None

    ltgt = []
    if libtgt is None:
        pass
    elif isinstance(libtgt, list):
        for l in libtgt:
            ltgt.append(install_shared_library(l))
    else:
        ltgt.append(install_shared_library(libtgt))

    deps = []
    if libtgt is None:
        pass
    elif isinstance(libtgt, list):
        for l in libtgt:
            deps.append(l)
    else:
        deps.append(libtgt)

    htgt, rawtgts = install_headers(
        tgtdir=hroot, srcdir=headers, patterns=headers_patterns, adddeps=deps)

    tgts = [htgt] + ltgt if libtgt else [htgt]

    if uninstall:
        # Add uninstall target as fake makefile
        # with weak targets binding.
        # makefile routine changed to 'clean'
        # TODO: Create explicit Uninstall target in .make.py
        default_core().add(
            licant.make.MakeFileTarget(
                tgt=uninstall,
                build=licant.make.MakeFileTarget.clean,
                makefile=licant.make.MakeFileTarget.clean,
                deps=[],
                weakdeps=rawtgts + ltgt
            )
        )

    return licant.fileset(tgt=tgt, targets=tgts)



# coding: utf-8

from argparse import ArgumentParser
from argparse import RawTextHelpFormatter

import licant.util
import licant.core

import sys
import os

default_target = None

parser = ArgumentParser(formatter_class=RawTextHelpFormatter)
opts = None
args = None


def make_freeargs_help():
    return "freeargs help TODO"


parser.add_argument(
    "-d",
    "--debug",
    action="store_true",
    default=False,
    help="print full system commands",
)
parser.add_argument(
    "-t",
    "--trace",
    action="store_true",
    default=False,
    help="print trace information",
)
parser.add_argument(
    "-j", "--threads", default='j', help="amount of threads for executor"
)
parser.add_argument(
    "-q",
    "--quite",
    action="store_true",
    default=False,
    help="don`t print shell operations",
)
parser.add_argument("freeargs", type=str, nargs="*", help=make_freeargs_help())

parser.add_argument("--printruntime", action="store_true", default=False)


def add_argument(*args, **kvargs):
    parser.add_argument(*args, **kvargs)


def parse(argv=sys.argv[1:]):
    global opts, args

    if not opts:
        opts = parser.parse_args(argv)
        args = opts.freeargs

    return opts, args


def execute_with_default_action(target, args):
    if not hasattr(target, "default_action"):
        licant.util.error(
            "target {} hasn't default_action (actions: {})".format(
                licant.util.yellow(target.tgt), licant.util.get_actions(target)
            )
        )
    return target.invoke(target.default_action, args=args, critical=True)


def __cliexecute(args, default, core):
    global default_target
    default_target = default

    if len(args) == 0:
        if default is None:
            licant.util.error("default target isn't set")

        target = core.get(default)
        return execute_with_default_action(target, [])

    fnd = args[0]

    # Try look up fnd in targets
    if core.has(fnd):
        target = core.get(fnd)

        if len(args) == 1 or isinstance(target, licant.core.Routine):
            return execute_with_default_action(target, args[1:])

        act = args[1]

        if not target.hasaction(act):
            licant.util.error(
                "{} is not action of target {}".format(
                    licant.util.yellow(act), licant.util.yellow(fnd)
                )
            )

        return target.invoke(act, *args[2:], critical=True)

    # Try look up fnd in actions of default_target
    if default is not None:
        dtarget = core.get(default)
        if dtarget.hasaction(fnd):
            return dtarget.invoke(fnd, *args[1:], critical=True)

    # Can't look fnd.
    licant.util.error(
        "Can't find routine "
        + licant.util.yellow(fnd)
        + ". Enough target or default target's action with same name."
    )


def cliexecute(
        default=None,
        colorwrap=False,
        argv=sys.argv[1:],
        core=licant.core.default_core()):
    if colorwrap:
        print(licant.util.green("[start]"))

    opts, args = parse(argv)

    core.runtime["debug"] = opts.debug or opts.trace
    core.runtime["trace"] = opts.trace
    core.runtime["quite"] = opts.quite

    cpu_count = os.cpu_count()
    core.runtime["threads"] = cpu_count if opts.threads == "j" else int(
        opts.threads)

    if opts.printruntime:
        print("PRINT RUNTIME:", core.runtime)

    try:
        __cliexecute(args, default=default, core=core)
    except KeyboardInterrupt:
        licant.util.error("interrupted by user")

    if colorwrap:
        print(licant.util.yellow("[finish]"))



# coding: utf-8

from __future__ import print_function

import licant
from licant.core import UpdatableTarget, Core, default_core
from licant.cache import fcache
from licant.util import purple, quite
import threading
import os
import time
import subprocess
import fcntl
import sys
import traceback

_rlock = threading.RLock()


def do_execute(target, rule, msgfield, prefix=None):
    core = target.core

    def sprint(*args, **kwargs):
        """print for multithread build"""
        try:
            with _rlock:
                print(*args, **kwargs)
        except Exception as e:
            print("Licant: Print error:", e)

    rule = rule.format(**target.__dict__)

    message = getattr(target, msgfield, None)

    if not core.runtime["quite"]:
        if not core.runtime["debug"] and message is not None:
            if not isinstance(message, quite):
                if prefix is not None:
                    sprint(prefix, message.format(**target.__dict__))
                else:
                    sprint(message.format(**target.__dict__))

        else:
            sprint(rule)

    proc = subprocess.Popen(
        rule, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout = proc.stdout
    stderr = proc.stderr
    ret = proc.wait()

    sprint(stdout.read().decode("utf-8"), end="")
    sys.stdout.flush()
    sprint(stderr.read().decode("utf-8"), end="")
    sys.stderr.flush()

    if target.is_file():
        target.update_info()

    return True if ret == 0 else False


class Executor:
    def __init__(self, rule, msgfield="message"):
        self.rule = rule
        self.msgfield = msgfield

    def __call__(self, target, **kwargs):
        return do_execute(target, self.rule, self.msgfield, **kwargs)


class DirectoryKeeper():
    def __init__(self, msgfield="message"):
        self.msgfield = msgfield

    def __call__(self, target, **kwargs):
        print("MAKEDIRS", target.tgt)
        if not target.exists():
            os.makedirs(target.tgt)
            target.update_info()


class MakeFileTarget(UpdatableTarget):
    __actions__ = UpdatableTarget.__actions__.union(
        {"actlist", "makefile", "clean"})

    def __init__(self, tgt, deps, **kwargs):
        UpdatableTarget.__init__(self, tgt, deps, **kwargs)
        self.default_action = "makefile"

    def clean(self):
        stree = self.core.depends_as_set(self.tgt, incroot=True)
        stree = [self.core.get(s) for s in stree]

        for target in stree:
            if if_file_and_exist(target) and target.clr:
                target.clr()

    def makefile(self):
        return self.recurse_update()

    def is_file(self):
        return True

    def mtime(self):
        """fileset mtime"""
        maxtime = 0
        for d in self.deps:
            dtgt = self.core.get(d)
            mtime = dtgt.mtime()
            if mtime is not None and mtime > maxtime:
                maxtime = mtime

        return maxtime


class FileTarget(MakeFileTarget):
    __actions__ = MakeFileTarget.__actions__.union(
        {"build", "clr"})

    def __init__(self, tgt, deps, force=False, use_dirkeep=True, **kwargs):

        tgt = licant.util.canonical_path(tgt)

        if use_dirkeep:
            dirpath = os.path.normpath(os.path.dirname(tgt))
            if not os.path.exists(dirpath):
                dirkeep(dirpath)
                deps = deps + [dirpath]

        MakeFileTarget.__init__(self, tgt, deps, **kwargs)
        self.isfile = True
        self.force = force
        self.clrmsg = "DELETE {tgt}"

    def update_info(self):
        fcache.update_info(self.tgt)

    def mtime(self):
        curinfo = fcache.get_info(self.tgt)
        return curinfo.mtime

    def is_exist(self):
        curinfo = fcache.get_info(self.tgt)
        return curinfo.exist

    def exists(self):
        return self.is_exist()

    def warn_if_not_exist(self):
        """Print warn if file isn't exist"""
        info = fcache.get_info(self.tgt)
        if not info.exist:
            print("Warn: file {} isn`t exist".format(purple(self.tgt)))

    def clr(self):
        """Delete this file."""
        do_execute(self, "rm -f {tgt}", "clrmsg")

    def internal_need_if(self):
        if self.force or not self.is_exist():
            return True

        maxtime = 0
        for dep in self.get_deplist():
            depmtime = dep.mtime()
            if dep.is_file() and depmtime > maxtime:
                maxtime = depmtime

        if maxtime > self.mtime():
            return True

        return False

    def update(self):
        sts = self.build(self)
        self.update_info()
        return sts

    def __lt__(self, other):
        return str(self) < str(other)

    def __gt__(self, other):
        return str(self) > str(other)


class DirectoryTarget(FileTarget):
    def internal_need_if(self):
        if not self.is_exist():
            return True
        return False

    def mtime(self):
        return 0

    def update_if_need(self):
        if self.internal_need_if():
            return self.update()
        else:
            return True

    def clr(self):
        """Prevent directory deletion."""
        pass


def source(tgt, deps=[]):
    """Index source file by licant core."""

    if default_core().exist(tgt):
        return

    target = FileTarget(
        build=lambda self: self.warn_if_not_exist(), deps=deps, tgt=tgt)
    target.clr = None
    target.dirkeep = licant.util.do_nothing
    return default_core().add(target)


def dirkeep(dirpath, message="MKDIR {tgt}"):
    """Create directory tree for this file if needed."""
    base_directory_path = os.path.normpath(os.path.dirname(dirpath))
    return default_core().add(
        DirectoryTarget(
            tgt=dirpath,
            build=DirectoryKeeper(),
            use_dirkeep=not os.path.exists(base_directory_path),
            message=message,
            deps=[]
        )
    )


def makedir(dirpath, message="MKDIR {tgt}"):
    return dirkeep(dirpath, message)


def copy(tgt, src, adddeps=[], message="COPY {src} {tgt}", core=default_core()):
    """Make the file copy target."""
    src = os.path.expanduser(str(src))
    tgt = os.path.expanduser(str(tgt))

    source(src)
    core.add(
        FileTarget(
            tgt=tgt,
            build=Executor("cp {src} {tgt}"),
            src=src,
            deps=[src] + adddeps,
            message=message,
        )
    )
    return tgt


def makefile(tgt, deps, do, core=default_core(), **kwargs):
    """Makefile target."""
    return core.add(
        FileTarget(
            tgt=tgt,
            build=Executor(do),
            deps=deps,
            **kwargs
        )
    )


def fileset(tgt, targets, deps=[], core=default_core(), **kwargs):
    """Make a fileset."""
    core.add(MakeFileTarget(tgt=tgt, deps=deps + targets, core=core, **kwargs))
    return tgt


def if_file_and_exist(target):
    if not isinstance(target, FileTarget):
        return False

    curinfo = fcache.get_info(target.tgt)
    return curinfo.exist


class MakeCore(Core):
    def __init__(self, debug=False):
        super().__init__(debug=debug)

    def makedir(self, dirpath, deps, message="MKDIR {tgt}"):
        return self.add(
            DirectoryTarget(
                tgt=dirpath,
                build=DirectoryKeeper(),
                use_dirkeep=False,
                message=message,
                deps=deps
            )
        )

    def dirkeep(self, path):
        """Create directory tree for this file if needed."""
        base_directory_path = os.path.normpath(os.path.dirname(path))
        if not os.path.exists(base_directory_path):
            deps = self.dirkeep(base_directory_path)
            if not self.has(base_directory_path):
                self.makedir(base_directory_path, deps=deps)
            return [base_directory_path]
        else:
            return []

    def touch(self, out, content="", deps=[]):
        """Create file if not exist."""

        def create_file(self):
            with open(self.tgt, "w") as f:
                f.write(content)

        out = os.path.expanduser(str(out))
        dirdeps = self.dirkeep(out)
        return self.add(FileTarget(
            tgt=out,
            build=create_file,
            deps=deps + dirdeps,
            message="TOUCH {tgt}",
            use_dirkeep=False,
            content=content
        ))

    def copy(self, src, dst, deps=[], message="COPY {src} {tgt}"):
        """Make the file copy target."""
        src = os.path.expanduser(str(src))
        dst = os.path.expanduser(str(dst))
        dirdeps = self.dirkeep(dst)
        self.source(src)
        return self.add(FileTarget(
            tgt=dst,
            build=Executor("cp {src} {tgt}"),
            src=src,
            deps=[src] + dirdeps + deps,
            use_dirkeep=False,
            message=message,
        ))

    def source(self, src):
        """Index source file by licant core."""
        if self.has(str(src)):
            return self.get(str(src))

        src = os.path.expanduser(str(src))
        target = self.add(FileTarget(
            tgt=src,
            build=lambda self: self.warn_if_not_exist(),
            deps=[],
            use_dirkeep=False,
            message="SOURCE {tgt}"
        ))
        target.clr = None
        target.dirkeep = licant.util.do_nothing
        return target

    def ftarget(self, tgt, build=None, deps=[], exec=None, message="FTARGET {tgt}"):
        """Make the file target."""
        if exec is not None:
            build = licant.Executor(exec)

        if build is None:
            raise Exception("build action is None")

        tgt = os.path.expanduser(str(tgt))
        dirdeps = self.dirkeep(tgt)
        return self.add(FileTarget(
            tgt=tgt,
            build=build,
            deps=deps + dirdeps,
            use_dirkeep=False,
            message=message
        ))



import sys
import os
import licant.util
from licant.core import default_core


class ScriptQueue:
    def __init__(self):
        self.stack = [sys.argv[0]]
        licant.make.source(sys.argv[0])

    def __execute(self, path):
        if default_core().runtime["debug"]:
            print("SUBSCRIPT", path)
        licant.make.source(path)
        self.stack.append(path)
        try:
            exec(open(path).read(), globals())
        except Exception as e:
            licant.util.error_exception("Exception in file {}".format(path), e)
        self.stack.pop()

    def execute(self, path):
        path = os.path.join(self.curdir(), path)
        self.__execute(path)

    def last(self):
        return self.stack[-1]

    def curdir(self):
        return os.path.dirname(self.last())

    def execute_recursive(self, root, pattern, hide=None, debug=False):
        root = os.path.join(self.curdir(), root)
        flst = licant.util.find_recursive(root, pattern, hide, debug)
        for f in flst:
            self.__execute(f)


scriptq = ScriptQueue()



# coding: utf-8

from cmath import inf
from licant.modules import mlibrary, module
from licant.cxx_make import host_toolchain
from licant.core import default_core
from licant.util import red, yellow, cxx_read_depends

import os
import licant.util as gu
import licant.make

import glob
import queue


class solver:
    def __init__(self, type, merge, include, default):
        self.type = type
        self.merge = merge
        self.include = include
        self.default = default


def base(base, local, solver, tbase, tlocal):
    return base


def local(base, local, solver, tbase, tlocal):
    if local is None:
        return solver.default
    return local


def concat(base, local, solver, tbase, tlocal):
    if local is None:
        local = []
    if base is None:
        base = []
    return base + local


def strconcat(base, local, solver, tbase, tlocal):
    if local is None:
        return base
    return base + " " + local


def concat_add_srcdir(base, local, solver, tbase, tlocal):
    if local is None:
        local = []
    else:
        if "srcdir" in tlocal.opts:
            srcdir = tlocal.opts["srcdir"]
        else:
            srcdir = ""

        local = list(
            map(
                lambda p: os.path.join(
                    tlocal.opts["__dir__"], srcdir, os.path.expanduser(p)
                ),
                local,
            )
        )
    # print(base + local)
    return base + local


def local_add_srcdir(base, local, solver, tbase, tlocal):
    if "srcdir" in tlocal.opts:
        srcdir = tlocal.opts["srcdir"]
    else:
        srcdir = ""
    if isinstance(local, type("str")):
        local = [local]
    if local is None:
        return []
    return list(
        map(
            lambda p: os.path.join(
                tlocal.opts["__dir__"], srcdir, os.path.expanduser(p)
            ),
            local,
        )
    )


def concat_add_locdir(base, local, solver, tbase, tlocal):
    if local is None:
        return base

    if isinstance(local, type("str")):
        local = [local]
    local = list(
        map(
            lambda p: os.path.join(
                tlocal.opts["__dir__"], os.path.expanduser(p)), local
        )
    )
    return base + local


def concat_add_locdir_second(base, local, solver, tbase, tlocal):
    """ Добавить локальный путь ко всем вторым элементам локального массива.
    Для поддержки локальных заголовков."""
    if local is None:
        return base
    local = [(pl[0], os.path.join(tlocal.opts["__dir__"], pl[1]))
             for pl in local]
    return base + local


def local_if_exist(base, local, solver, tbase, tlocal):
    return base if local is None else local


def concat_to_submodule(base, local, solver, tbase, tlocal):
    if local is None:
        local = []
    if base is None:
        base = []
    lm = [licant.modules.submodule(l) for l in local]
    return base + lm


cxx_module_field_list = {
    #                                       # merge                     #include                    #default
    "srcdir": solver("str", local, base, "."),
    "depends": solver("list", local, base, []),
    "objects": solver("list", local_add_srcdir, concat_add_srcdir, []),
    "sources": solver("list", local_add_srcdir, concat_add_srcdir, []),
    "qt_moc": solver("list", local_add_srcdir, concat_add_srcdir, []),
    "qt_ui": solver("list", local_add_srcdir, concat_add_srcdir, []),
    "libs": solver("list", concat, concat, []),
    "libdirs": solver("list", concat, concat, []),
    "target": solver("str", local_if_exist, base, "target"),
    "include_paths": solver("list", concat_add_locdir, concat_add_locdir, []),
    "cxxstd": solver("str", local_if_exist, local_if_exist, "c++17"),
    "ccstd": solver("str", local_if_exist, local_if_exist, "c11"),
    "cxx_flags": solver("str", strconcat, strconcat, ""),
    "cc_flags": solver("str", strconcat, strconcat, ""),
    "ld_flags": solver("str", strconcat, strconcat, ""),
    "modules": solver("list", local, concat, []),
    "type": solver("str", local, base, "objects"),
    "builddir": solver("str", local_if_exist, base, "build"),
    "toolchain": solver("toolchain", local_if_exist, base, host_toolchain),
    "optimize": solver("str", local_if_exist, base, ""),
    "include_modules": solver("list", concat_to_submodule, base, []),
    "defines": solver("list", concat, concat, []),
    "ldscripts": solver("list", concat_add_locdir, concat_add_locdir, []),
    "local_headers": solver(
        "list", concat_add_locdir_second, concat_add_locdir_second, []
    ),
    "mdepends": solver("list", local, base, []),
    "debug": solver("str", local_if_exist, base, "-g"),
}


class CXXModuleOptions:
    def __getitem__(self, key):
        return self.opts[key]

    def check(self):
        for k, v in self.opts.items():
            if k not in licant.modules.special:
                if k not in self.table:
                    print("Unresolved option: {}".format(red(k)))
                    print("cxx_module_field_list:", cxx_module_field_list)
                    exit(-1)

                if v.__class__.__name__ != self.table[k].type:
                    print(
                        "Option should be object of {}: {}".format(
                            yellow(self.table[k].type), red(k)
                        )
                    )
                    exit(-1)

    def set_default_if_empty(self, table=cxx_module_field_list):
        for k in table:
            if k not in self.opts:
                self.opts[k] = table[k].default

    def merge(self, other, fnc):
        resopts = {}
        for k, v in self.opts.items():

            base = self.opts[k]
            if k in other.opts:
                local = other.opts[k]
            else:
                local = None

            func = getattr(self.table[k], fnc)
            resopts[k] = func(base, local, self.table[k], self, other)

        return CXXModuleOptions(**resopts)

    def __init__(self, table=cxx_module_field_list, **kwargs):
        self.opts = kwargs
        self.table = table
        self.check()

    def __str__(self):
        return str(self.opts)


def cxx_options_from_modopts(modopts):
    """Конвертировать информацию из модуля в структуру опций cxx_make"""

    cxxstd = "-std=" + modopts["cxxstd"]
    ccstd = "-std=" + modopts["ccstd"]

    cxx_flags = " ".join([cxxstd, modopts["debug"], modopts["cxx_flags"]])
    cc_flags = " ".join([ccstd, modopts["debug"], modopts["cc_flags"]])

    ld_srcs_add = "".join([" -L" + os.path.expanduser(l)
                           for l in modopts["libdirs"]])
    ld_srcs_add += "".join([" -l" + l for l in modopts["libs"]])

    return licant.cxx_make.options(
        toolchain=modopts["toolchain"],
        include_paths=modopts["include_paths"],
        cc_flags=cc_flags,
        cxx_flags=cxx_flags,
        defines=modopts["defines"],
        ldscripts=modopts["ldscripts"],
        ld_flags=modopts["ld_flags"],
        ld_srcs_add=ld_srcs_add,
        optimize=modopts["optimize"],
    )


def build_paths(srcs, opts, ext, prefix=None, builddir=None):
    if builddir is None:
        builddir = opts.opts["builddir"]

    objs = []

    for s in srcs:
        if prefix:
            split = os.path.split(s)
            s = os.path.join(*split[:-1], prefix + split[-1])
        if os.path.isabs(s):
            s = builddir + gu.changeext(s, ext)
            objs.append(s)
        else:
            objs.append(
                os.path.normpath(
                    os.path.join(
                        builddir, gu.changeext(s, ext).replace("..", "__")
                    )
                )
            )

    return objs


def sources_paths(opts):
    ret = []
    for s in opts["sources"]:
        if "*" in s:
            ret.extend(glob.glob(s))
        else:
            ret.append(s)
    return ret


def qt_moc_paths(opts):
    ret = []
    for s in opts["qt_moc"]:
        if "*" in s:
            ret.extend(glob.glob(s))
        else:
            ret.append(s)
    return ret


def qt_ui_paths(opts):
    ret = []
    for s in opts["qt_ui"]:
        if "*" in s:
            ret.extend(glob.glob(s))
        else:
            ret.append(s)
    return ret


def link_objects(srcs, objs, deps, cxxopts, adddeps):

    for s, o, d in zip(srcs, objs, deps):
        if not default_core().has(s):
            licant.make.source(s)

        headers = cxx_read_depends(d)
        force = headers is None
        if headers is None:
            headers = []
        else:
            for h in headers:
                if not default_core().has(h):
                    licant.make.source(h)

        licant.cxx_make.depend(
            src=s, tgt=d, opts=cxxopts, deps=[s] + adddeps + headers, force=force
        )
        licant.cxx_make.object(
            src=s, tgt=o, opts=cxxopts, deps=[s, d] + adddeps + headers
        )


def link_qt_moc(mocs, srcs, cxxopts, adddeps):
    for m, s in zip(mocs, srcs):
        licant.make.source(m)
        licant.cxx_make.qt_moc(src=m, tgt=s, opts=cxxopts, deps=[m] + adddeps)


def link_qt_ui(uis, srcs, cxxopts, adddeps):
    for m, s in zip(uis, srcs):
        licant.make.source(m)
        licant.cxx_make.qt_uic(src=m, tgt=s, opts=cxxopts, deps=[m] + adddeps)


def executable(srcs, opts):
    cxxopts = cxx_options_from_modopts(opts)
    return licant.cxx_make.executable(tgt=opts["target"], srcs=srcs, opts=cxxopts)


def dynlib(srcs, opts):
    cxxopts = cxx_options_from_modopts(opts)
    return licant.cxx_make.dynamic_library(
        tgt=opts["target"], srcs=srcs, opts=cxxopts)


def statlib(srcs, opts):
    cxxopts = cxx_options_from_modopts(opts)
    return licant.cxx_make.static_library(tgt=opts["target"], srcs=srcs, opts=cxxopts)


def virtual(srcs, opts):
    licant.core.target(tgt=opts["target"], deps=srcs)
    return opts["target"]


def collect_modules(mod):
    # Альтернативная система модулей.
    mdepends_default = queue.Queue()
    mdepends = dict()
    included_from = dict()

    class SortKey:
        def __init__(self):
            self.sortkey = 0

        def __call__(self, mod):
            mod.__sortkey = self.sortkey
            self.sortkey += 1

    setsortkey = SortKey()

    def add_included_from(name, incfrom):
        if name not in included_from:
            included_from[name] = set()

        included_from[name].add(incfrom)

    def specify_implementation(name, mod, incfrom):
        mdepends[name] = mod
        setsortkey(mod)
        add_included_from(name, incfrom)

    def request_implementation(name, incfrom):
        mdepends_default.put(name)
        add_included_from(name, incfrom)

    def collect_modules(mod):
        if default_core().runtime["trace"]:
            print("trace: collect_modules( {} )".format(mod.name))

        for md in mod.opts["mdepends"]:
            if isinstance(md, str):
                if md in mdepends:
                    continue

                request_implementation(name=md, incfrom=mod.name)

            elif isinstance(md, tuple):
                nmod = mlibrary.get(md[0], md[1])

                if md[0] in mdepends and mdepends[md[0]] is not nmod:
                    licant.error(
                        "This module added early and it's different: mod:{} now:{} early:{}".format(
                            licant.util.yellow(nmod.name),
                            licant.util.yellow(nmod.impl),
                            licant.util.yellow(mdepends[md[0]].impl),
                        )
                    )

                specify_implementation(name=md[0], mod=nmod, incfrom=mod.name)

                if "mdepends" in nmod.opts:
                    # Сразу же выполняем обход по данному модулю
                    collect_modules(nmod)

            else:
                licant.error("Very strange mpdepend")

    if "mdepends" in mod.opts:
        collect_modules(mod)

    while not mdepends_default.empty():
        name = mdepends_default.get()
        if name in mdepends:
            continue

        defmod = mlibrary.get_default(name)
        specify_implementation(name, defmod, mod.name)

        if "mdepends" in defmod.opts:
            collect_modules(defmod)

    for m in mdepends:
        mdepends[m].included_from = included_from[m]

    return [
        mdepends[x]
        for x in sorted(mdepends, key=lambda x: mdepends[x]._SortKey__sortkey)
    ]


def prepare_targets(name, impl=None, core=licant.core.default_core(), **kwargs):
    opts = CXXModuleOptions(**kwargs)
    opts.set_default_if_empty()

    def modmake(name, impl, baseopts):
        mod = mlibrary.get(name, impl)

        modopts = CXXModuleOptions(**mod.opts)
        locopts = baseopts.merge(modopts, "merge")

        adddeps = []  # mod.stack

        # Проходим по include_modules, рекурсивно формируем локальные опции.
        def include_modules(locopts, lst):
            retopts = locopts
            for simod in lst:
                imod = mlibrary.get(simod.name, simod.impl)
                retopts = retopts.merge(imod, "include")

                if "include_modules" in imod.opts:
                    retopts = include_modules(
                        retopts, imod.opts["include_modules"])

            return retopts

        locopts = include_modules(locopts, locopts["include_modules"])

        # Система модулей mdepends
        for imod in collect_modules(mod):
            locopts = locopts.merge(imod, "include")

        local_headers_targets = []
        if len(locopts["local_headers"]):
            locopts["include_paths"].append(
                locopts["builddir"] + "/__LOCAL_HEADERS__/")
            for pair in locopts["local_headers"]:
                licant.make.source(pair[1])
                tgtpath = os.path.join(
                    locopts["builddir"], "__LOCAL_HEADERS__", pair[0]
                )
                licant.make.copy(src=pair[1], tgt=tgtpath)

                local_headers_targets.append(tgtpath)

        licant.make.fileset(name + "__local_headers__", local_headers_targets)
        adddeps.append(name + "__local_headers__")

        for dep in locopts["depends"]:
            if "*" in dep:
                adddeps.extend(glob.glob(dep))
            else:
                adddeps.append(dep)

        locsrcs = sources_paths(locopts)
        locobjs = build_paths(locsrcs, locopts, "o")
        locdeps = build_paths(locsrcs, locopts, "d")

        cxxopts = cxx_options_from_modopts(locopts)

        if len(locopts["qt_ui"]) != 0:
            uidir = os.path.join(locopts.opts["builddir"], "ui")
            locui = qt_ui_paths(locopts)
            locuihxx = build_paths(locui, locopts, "h",
                                   builddir=uidir,
                                   prefix="ui_")
            link_qt_ui(locui, locuihxx, cxxopts, adddeps)
            adddeps.extend(locuihxx)

            if uidir not in locopts.opts["include_paths"]:
                locopts.opts["include_paths"].append(uidir)

            # Recompile with ui include path
            cxxopts = cxx_options_from_modopts(locopts)

        if len(locopts["qt_moc"]) != 0:
            locmoc = qt_moc_paths(locopts)
            locmoccxx = build_paths(locmoc, locopts, "h.cxx")
            locmocobjs = build_paths(locmoc, locopts, "h.cxx.o")
            locmocdeps = build_paths(locmoc, locopts, "h.cxx.d")
            link_qt_moc(locmoc, locmoccxx, cxxopts, adddeps)

            locsrcs.extend(locmoccxx)
            locobjs.extend(locmocobjs)
            locdeps.extend(locmocdeps)

        link_objects(locsrcs, locobjs, locdeps, cxxopts, adddeps)

        if len(locopts["objects"]) != 0:
            for t in locopts["objects"]:
                licant.make.source(t)
            locobjs.extend(locopts["objects"])

        submodules_results = []
        # print(locopts["modules"])
        for smod in locopts["modules"]:
            submodules_results += modmake(smod.name, smod.impl, locopts)

        if locopts["type"] == "application":
            return ([executable(locobjs + submodules_results, locopts)], locopts)
        elif locopts["type"] == "shared_library":
            return ([dynlib(locobjs + submodules_results, locopts)], locopts)
        elif locopts["type"] == "static_library":
            return ([statlib(locobjs + submodules_results, locopts)], locopts)
        elif locopts["type"] == "objects":
            return (locobjs + submodules_results, locopts)
        else:
            print("Wrong type of assemble: {}".format(gu.red(locopts["type"])))
            exit(-1)

    res, locopts = modmake(name, impl, opts)
    return (res, locopts)


def task(name, target, impl, type, core, **kwargs):
    if type != "objects":
        if target is None or target == name:
            target = name
            module_name = name + "__" + "module"
        else:
            module_name = name + "__" + "module"
        licant.modules.module(module_name, impl=impl, type=type,
                              target=target, **kwargs)
    else:
        licant.modules.module(name, impl=impl, type=type, **kwargs)
    ret, opts = prepare_targets(module_name, core=core)
    #licant.make.fileset(tgt=name, targets=ret, finalopts=opts, core=core)

    if type != "objects":
        return ret[0]
    else:
        return licant.make.fileset(tgt=name, targets=ret, finalopts=opts, core=core)


def application(name, target=None, impl=None, type="application", core=licant.core.default_core(), **kwargs):
    return task(name, target, impl, type, core=core, **kwargs)


def shared_library(name, target=None, impl=None, type="shared_library", core=licant.core.default_core(), **kwargs):
    return task(name, target, impl, type, core=core, **kwargs)


def static_library(name, target=None, impl=None, type="static_library", core=licant.core.default_core(), **kwargs):
    return task(name, target, impl, type, core=core, **kwargs)


def objects(name, target=None, impl=None, type="objects", **kwargs):
    return task(name, target, impl, type, **kwargs)


def library(*args, shared=True, **kwargs):
    if shared:
        return shared_library(*args, **kwargs)
    else:
        return static_library(*args, **kwargs)


def static_and_shared(name, static_lib, shared_lib, **kwargs):
    a = static_library(static_lib, **kwargs)
    b = shared_library(shared_lib, **kwargs)
    licant.make.fileset(name, [a.tgt, b.tgt])
    return a, b


def print_collect_list(target, *args):
    if len(args) > 0:
        name = args[0]
    else:
        print("You should specify target name")
        return
    for m in sorted(collect_modules(mlibrary.get(name)), key=lambda x: x.name):
        if hasattr(m, "impl"):
            print("{}:{}".format(m.name, m.impl))
        else:
            print(m.name)


def print_finalopts(target, *args):
    if len(args) > 0:
        name = args[0]
    else:
        name = licant.cli.default_target
    print(licant.core.core.get(name).finalopts)


def included_from(target, *args):
    if len(args) > 0:
        name = args[0]
    else:
        print("You should specify target name")
        return

    for m in sorted(collect_modules(mlibrary.get(name)), key=lambda x: x.name):
        print(m.name, ":", m.included_from)


modules_target = licant.core.Target(
    tgt="cxxm",
    deps=[],
    collect_modules=print_collect_list,
    included_from=included_from,
    finalopts=print_finalopts,
    actions={"collect_modules", "finalopts", "included_from"},
    __help__="Info about collected modules",
)

licant.core.default_core().add(modules_target)



#!/usr/bin/python
#coding: utf-8

import sys
import os
import json
from optparse import OptionParser

import licant.libs

paths = {}


def add_record(name, path):
    global paths
    abspath = os.path.abspath(path)
    exists = os.path.exists(abspath)

    if not exists:
        print("File {} is not exists".format(path))
        exit(-1)

    paths[name] = abspath


def licant_libs_utility():
    global paths
    usage = "usage: %prog [options] lib path"
    parser = OptionParser(usage=usage)
    parser.add_option("-u", "--user", action="store_true", default=False)
    parser.add_option("-r", "--remove", action="store_true", default=False)
    parser.add_option("-l", "--list", action="store_true", default=False)
    opts, args = parser.parse_args(sys.argv[1:])

    paths_file = licant.libs.lpath if opts.user else licant.libs.gpath

    if (os.path.exists(paths_file)):
        try:
            paths = json.load(open(paths_file))
        except:
            print("Load error:")
            print("Wrong format", paths_file)
            exit(-1)
    else:
        paths = {}

    if opts.list:
        print("Libraries:")
        for k, v in paths.items():
            print(k, v)
        exit(0)

    if len(args) > 2 or (opts.remove and len(args) != 1):
        parser.error("incorrect numbers of argument")
        exit(0)

    if len(args) == 0:
        parser.help()
        exit(0)

    path = None
    name = None
    if len(args) == 2 and opts.remove == False:
        name = args[0]
        path = args[1]
        add_record(name, path)

    if len(args) == 1 and opts.remove == False:
        import glob
        ret = glob.glob(os.path.join(args[0], "*.g.py"))
        for f in ret:
            fname = os.path.basename(f)
            name = fname.split('.')[0]
            path = f
            add_record(name, path)

    if opts.remove:
        try:
            del paths[name]
        except:
            print("Unregistred library {}".format(name))

    try:
        print(f"update: {paths_file}, text: {json.dumps(paths, indent=4)}")
        json.dump(paths, open(paths_file, "w"))
    except IOError as e:
        print("Dump error:")
        print(e)
        exit(-1)


if __name__ == "__main__":
    licant_libs_utility()



from concurrent.futures import thread
from threading import Thread
from queue import Queue
import queue
import threading
import sys
from licant.util import invert_depends_dictionary, red


class DependableTarget:
    def __init__(self, name, deps, what_to_do, args=[], kwargs={}):
        self.name = name
        self.deps = set(deps)
        self.what_to_do = what_to_do
        self.args = args
        self.kwargs = kwargs
        self._is_done = False

    def doit(self):
        result = self.what_to_do(*self.args, **self.kwargs)
        self._is_done = True
        return result

    def is_done(self):
        return self._is_done


class DependableTargetRuntime:
    def __init__(self, deptarget, task_invoker):
        self.deptarget = deptarget
        self.depcount = len(deptarget.deps)
        self.inverse_deps = set()
        self.task_invoker = task_invoker

    def is_done(self):
        return self.deptarget.is_done()

    def set_inverse_deps(self, inverse_deps: set):
        self.inverse_deps = inverse_deps

    def deps(self):
        return self.deptarget.deps

    def decrease_inverse_deps_count(self):
        self.depcount -= 1
        if self.depcount == 0:
            self.task_invoker.add_target(self)
        assert self.depcount >= 0

    def doit(self):
        result = self.deptarget.doit()
        with self.task_invoker.mtx:
            for dep in self.inverse_deps:
                dep.decrease_inverse_deps_count()
        return result

    def count_of_deps(self):
        return len(self.deptarget.deps)

    def name(self):
        return self.deptarget.name

    def __str__(self) -> str:
        return self.name()

    def __repr__(self) -> str:
        return self.name()


class TaskInvoker:
    def __init__(self, threads_count: int, trace=False):
        self.queue = Queue()
        self.threads_count = threads_count
        self.threads = []
        self.thread_on_base = [True] * threads_count
        self.done = False
        self.mtx = threading.Lock()
        self.trace = trace
        self.error_while_execution = False

    def start(self):
        if self.threads_count == 1:
            if self.trace:
                print("[Trace] single thread mode")
            self.single_worker()
            return

        if self.trace:
            print(f"[Trace] start with {self.threads_count} threads")

        for i in range(self.threads_count):
            t = Thread(target=self.worker, args=(i,))
            t.start()
            self.threads.append(t)

    def single_worker(self):
        while not self.queue.empty():
            task = self.queue.get()
            if self.trace:
                print(f"[Trace] do: {task.name()}")
            result = task.doit()
            if self.trace:
                print(f"[Trace] result of last task: ", result)
            if result is False:
                self.done = True
                self.error_while_execution = True
                print(
                    f"{red('LicantError')}: Error while executing task {task.name()}")
                break

    def worker(self, no):
        while not self.done:
            try:
                task = self.queue.get(timeout=0.4)
                with self.mtx:
                    self.thread_on_base[no] = False
                if self.trace:
                    print(f"[Trace] thread:{no} do task: {task.name()}")
                result = task.doit()

                if self.done:
                    break

                if self.trace:
                    print(f"[Trace] thread:{no} result of last task: ", result)
                if result is False:
                    self.done = True
                    self.error_while_execution = True
                    print(
                        f"{red('LicantError')}: Error while executing task {task.name()}")
                    break

                with self.mtx:
                    self.thread_on_base[no] = True
                    if all(self.thread_on_base) and self.queue.empty():
                        self.done = True
                        break
            except queue.Empty:
                if all(self.thread_on_base) and self.queue.empty():
                    self.done = True
                    break
            except KeyboardInterrupt:
                self.done = True
                self.error_while_execution = True
                break

    def add_target(self, target):
        self.queue.put(target)

    def stop(self, wait=True):
        self.done = True
        if wait:
            for t in self.threads:
                t.join()

    def wait(self):
        try:
            for t in self.threads:
                t.join()
        except KeyboardInterrupt:
            print(f"{red('LicantError')}: Execution was interrupted by user")
            self.done = True
            self.error_while_execution = True
            for t in self.threads:
                t.join()


class UnknowTargetError(Exception):
    pass


class NoOneNonDependableTarget(Exception):
    pass


class CircularDependencyError(Exception):
    def __init__(self, lst):
        self.lst = lst
        Exception.__init__(self, lst)


class DoubleDependsError(Exception):
    def __init__(self, dep):
        Exception.__init__(self, dep)


class ConnectivityError(Exception):
    def __init__(self, nonvisited):
        self.lst = nonvisited
        Exception.__init__(self, self.lst)


class InverseRecursiveSolver:
    def __init__(self, targets: list, count_of_threads: int = 1, trace: bool = False):
        self.trace = trace
        self.check(targets)
        self.double_depends_check(targets)
        self.task_invoker = TaskInvoker(count_of_threads, trace)

        self.deptargets = [DependableTargetRuntime(
            target, self.task_invoker) for target in targets]

        self.names_to_deptargets = {
            target.name(): target for target in self.deptargets}
        deps_of_targets = self.collect_depends_of_targets(self.deptargets,
                                                          self.names_to_deptargets)
        inverse_deps_of_targets = invert_depends_dictionary(deps_of_targets)

        assert len(inverse_deps_of_targets) == len(self.deptargets)

        for deptarget in self.deptargets:
            deptarget.set_inverse_deps(inverse_deps_of_targets[deptarget])

        non_dependable_targets = self.get_non_dependable_targets()
        for target in non_dependable_targets:
            self.task_invoker.add_target(target)

        if len(non_dependable_targets) == 0:
            raise NoOneNonDependableTarget()

        self.connectivity_check(self.deptargets, non_dependable_targets)

    def double_depends_check(self, targets):
        for target in targets:
            for dep in target.deps:
                count = 0
                for dep2 in target.deps:
                    if dep == dep2:
                        count += 1
                if count > 1:
                    raise DoubleDependsError(dep)

    def dfs(self, target, visited, path):
        visited.add(target)
        path.append(target)
        for dep in target.inverse_deps:
            if dep in path:
                raise CircularDependencyError(path + [dep])
            if dep not in visited:
                self.dfs(dep, visited, path)
        path.pop()

    def connectivity_check(self, deptargets, non_dependable_targets):
        visited = set()
        path = []
        for target in non_dependable_targets:
            self.dfs(target, visited, path)

        if len(visited) != len(deptargets):
            nonvisited = set(deptargets) - visited
            raise ConnectivityError(nonvisited)

    def collect_depends_of_targets(self, deptargets, names_to_deptargets):
        try:
            deps_of_targets = {}
            for deptarget in deptargets:
                deps_of_targets[deptarget] = set()
                for dep in deptarget.deps():
                    deps_of_targets[deptarget].add(names_to_deptargets[dep])
            return deps_of_targets
        except KeyError as e:
            raise UnknowTargetError(e)

    def check(self, targets):
        for target in targets:
            if not isinstance(target, DependableTarget):
                raise TypeError(
                    "Target must be DependableTarget, but:", target.__class__)
            for dep in target.deps:
                if not isinstance(dep, str):
                    raise TypeError("Dep must be str")

    def get_non_dependable_targets(self):
        return [target for target in self.deptargets if target.count_of_deps() == 0]

    def exec(self):
        self.task_invoker.start()
        self.task_invoker.wait()
        if not self.task_invoker.error_while_execution:
            assert self.task_invoker.queue.empty()
            assert all(d.depcount == 0 for d in self.deptargets)
            assert all(d.is_done() for d in self.deptargets)
        if self.trace:
            print("[Trace] Execution finished. Status:",
                  not self.task_invoker.error_while_execution)
        return not self.task_invoker.error_while_execution



import subprocess
import os
from licant.cli import cliexecute as ex
from licant.core import Core, Target, UpdatableTarget
from licant.core import default_core
from licant.core import routine_decorator as routine
from licant.core import do
from licant.core import get_target
from licant.make import copy, fileset, Executor, makefile, source, makedir, MakeCore, DirectoryTarget, FileTarget
from licant.cxx_make import objcopy
from licant.cxx_modules import application as cxx_application
from licant.cxx_modules import shared_library as cxx_shared_library
from licant.cxx_modules import static_library as cxx_static_library
from licant.cxx_modules import static_and_shared as cxx_static_and_shared
from licant.cxx_modules import library as cxx_library
from licant.cxx_modules import objects as cxx_objects
from licant.modules import module, implementation, submodule
from licant.modules import module_default_implementation as module_defimpl
from licant.modules import module_default_implementation
from licant.util import error
from licant.cxx_make import gcc_toolchain, clang_toolchain, host_toolchain
import licant.scripter
from licant.libs import include

__version__ = "1.13.6"


def directory():
    return licant.scripter.scriptq.curdir()


def execute(path):
    licant.scripter.scriptq.execute(path)


def execute_recursive(*argv, **kwargs):
    licant.scripter.scriptq.execute_recursive(*argv, **kwargs)


def about():
    return "I'm Licant"


class Object(object):
    pass


glbfunc = Object()
attribute_store = Object()


def global_function(var):
    setattr(glbfunc, var.__name__, var)


def import_attribute(name, var):
    setattr(attribute_store, name, var)


def attribute(name):
    return getattr(attribute_store, name)

# also os.system but throw exception on error


def system(cmd, message=None):
    if message is not None:
        print(message)
    else:
        print(cmd)
    status = subprocess.check_call(cmd, shell=True)
    if status != 0:
        raise Exception("system error")


def mtime(path):
    return os.path.getmtime(path)


__all__ = [
    "include",
    "system",
    "error",
    "module_default_implementation",
    "module_defimpl",
    "module",
    "implementation",
    "submodule",
    "cxx_modules",
    "cxx_objects",
    "cxx_library",
    "cxx_application",
    "shared_library",
    "static_library",
    "static_and_shared",
    "makedir",
    "makefile",
    "fileset",
    "copy",
    "source",
    "do",
    "routine",
    "routine_decorator",
    "mtime",
    "UpdatableTarget",
    "Core",
    "core",
    "cxx_static_library",
    "cxx_shared_library",
    "objcopy",
    "get_target",
    "Target",
    "ex",
    "MakeCore",
    "DirectoryTarget",
    "FileTarget",
]



# coding: utf-8

import licant.util
import threading
import types
import inspect
from enum import Enum
import re
from functools import partial
import random
import glob
import os
from licant.solver import DependableTarget, InverseRecursiveSolver


class WrongAction(Exception):
    def __init__(self, obj, actname):
        self.obj = obj
        self.actname = actname

    def __str__(self):
        return "WrongAction: obj:{} actname:{} class:{} dict:{self.obj.__dict__}".format(self.obj, self.actname, self.obj.__class__, self.obj.__dict__)


class NoneDictionary(dict):
    def __init__(self):
        dict.__init__(self)

    def __getitem__(self, idx):
        try:
            return dict.__getitem__(self, idx)
        except Exception:
            return None


class Core:
    def __init__(self, debug=False):
        self._targets = {}
        self.help_showed_targets = []
        self.runtime = NoneDictionary()
        self.debug = debug
        self.depends_as_set_lazy_cache = {}

    def trace_mode(self):
        return self.runtime["trace"]

    def exist(self, name):
        return self.has(name)

    def add(self, target):
        """Add new target"""
        target.core = self
        self._targets[target.tgt] = target

        if self.debug:
            print("add target: " + target.tgt)

        if target.__help__ is not None:
            self.help_showed_targets.append(target)

        return target

    def get(self, tgt):
        """Get target object"""
        if str(tgt) in self._targets:
            return self._targets[str(tgt)]

        if licant.util.canonical_path(tgt) in self._targets:
            # access optimization becouse canonical_path is very slow
            self._targets[tgt] = self._targets[licant.util.canonical_path(tgt)]
            return self._targets[licant.util.canonical_path(tgt)]

        licant.util.error("unregistred target " + licant.util.yellow(tgt))

    def has(self, tgt):
        """Check if target exists"""
        if tgt in self._targets:
            return True

        if licant.util.canonical_path(tgt) in self._targets:
            # access optimization becouse canonical_path is very slow
            self._targets[tgt] = self._targets[licant.util.canonical_path(tgt)]
            return True

        return False

    def depends_as_set_impl(self, tgt, accum):
        target = self.get(str(tgt))
        for d in target.deps:
            if d not in accum:
                accum.add(d)
                self.depends_as_set_impl(d, accum)

    def depends_as_set(self, tgt, incroot):
        if tgt in self.depends_as_set_lazy_cache:
            return self.depends_as_set_lazy_cache[tgt]

        accumulator = set()
        if incroot:
            accumulator.add(str(tgt))
        self.depends_as_set_impl(tgt, accumulator)

        ret = sorted(accumulator)
        self.depends_as_set_lazy_cache[tgt] = ret
        return ret

    def target(self, name, deps=[], **kwargs):
        """Create new target"""
        return self.add(Target(name, deps=deps, **kwargs))

    def updtarget(self, name, deps=[], **kwargs):
        """Create new target"""
        return self.add(UpdatableTarget(name, deps=deps, **kwargs))

    def do(self, target, action=None, args=[], kwargs={}):
        """Do action on target"""
        if isinstance(target, str):
            target = self.get(target)

        if isinstance(target, (list, tuple)):
            for t in target:
                self.do(t, action, args, kwargs)
            return

        if action is None:
            action = target.default_action

        target.invoke(action, args=args, kwargs=kwargs)

    def routine_do(self, func=None, deps=[], update_if=lambda s: False, tgt=None):
        self.add(Routine(func=func, deps=deps, update_if=update_if, tgt=tgt))
        return func

    def routine(self, func=None, **kwargs):
        if inspect.isfunction(func):
            return self.routine_do(func, **kwargs)
        else:
            def decorator(func):
                return self.routine_do(func, **kwargs)
            return decorator


_default_core = Core()


def default_core():
    return _default_core


class Target:
    __actions__ = {"actlist", "print", "dependies"}

    def __init__(self, tgt, deps=[], action=lambda s: None, need_if=lambda s: True, weakdeps=[], actions=None, __help__=None, **kwargs):
        self.tgt = tgt
        deps = [self.to_name_if_needed(dep) for dep in deps]
        deps = self.expand_globs(deps)
        self.deps = deps
        self.need_if = need_if
        self.weakdeps = set(weakdeps)
        self.action = action
        self.default_action = "action"
        for k, v in kwargs.items():
            setattr(self, k, v)

        if actions is not None:
            self.__actions__ = self.__actions__.union(set(actions))

        self.__help__ = __help__

        self.need_by_self = None
        self.need_by_deps = None
        self.cached_deplist = None

    def to_name_if_needed(self, dep):
        if isinstance(dep, Target):
            return dep.tgt
        else:
            return dep

    def trace_mode(self):
        return self.core.trace_mode()

    def dependies(self):
        print(self.deps)

    def name(self):
        return self.tgt

    def is_file(self):
        return False

    def expand_globs(self, deps):
        import licant.make
        ret = []
        for d in deps:
            if "*" in d:
                ret.extend(glob.glob(d))
            else:
                ret.append(d)
        for r in ret:
            if os.path.exists(r):
                licant.make.source(r)
        return ret

    def action_if_need(self):
        need = self.need_if(self)
        self.need_by_self = need
        if need:
            self.action(self)

    def get_deplist(self):
        if self.cached_deplist is None:
            self.cached_deplist = [self.core.get(d) for d in self.deps]

        return self.cached_deplist

    def actlist(self):
        print(licant.util.get_actions(self))

    def print(self):
        print(self.__dict__)

    def hasaction(self, act):
        return act in self.__actions__

    def invoke(self, funcname: str, args=[], critical: bool = False, kwargs={}):
        """Invoke func function or method, or mthod with func name for this target

                Поддерживается несколько разных типов func.
                В качестве func может быть вызвана внешняя функция с параметром текущей цели,
                или название локального метода.
                critical -- Действует для строкового вызова. Если данный attr отсутствует у цели,
                то в зависимости от данного параметра может быть возвращен None или выброшено исключение.
                """
        if self.core.runtime["trace"]:
            print(
                "TRACE: Invoke: tgt:{}, act:{}, args:{}, kwargs:{}".format(
                    self.tgt, funcname, args, kwargs
                )
            )

        func = getattr(self, funcname, None)
        if func is None:
            if critical:
                print("wrong action: {}".format(funcname))
                raise WrongAction(self, funcname)
            return None

        if isinstance(func, types.MethodType):
            # return licant.util.cutinvoke(func, *args, **kwargs)
            return func(*args, **kwargs)

        else:
            # return licant.util.cutinvoke(func, self, *args, **kwargs)
            return func(self, *args, **kwargs)

    def __repr__(self):
        """По умолчанию вывод Target на печать возвращает идентификатор цели"""
        return self.tgt


class UpdatableTarget(Target):
    __actions__ = Target.__actions__.union(
        {"recurse_update", "recurse_update_get_args", "update", "update_if_need"}
    )

    def __init__(
        self,
        tgt,
        deps,
        need_if=None,
        default_action="recurse_update_get_args",
        ** kwargs
    ):
        Target.__init__(self, tgt, deps,
                        default_action=default_action, **kwargs)
        self.need_if = need_if

    def recurse_update_get_args(self):
        return self.recurse_update(threads=self.core.runtime["threads"])

    def update(self, *args, **kwargs):
        return True

    def recursive_update_needed_request(self):
        if self.need_to_update():
            return True

        for dep in self.get_deplist():
            if dep.recursive_update_needed_request():
                return True

        return False

    def has_updated_depends(self):
        alldeps = self.core.depends_as_set(self, incroot=False)
        alldeps = [self.core.get(d) for d in alldeps]

        for d in alldeps:
            if (
                not isinstance(d, UpdatableTarget)
            ):
                raise Exception("Nonupdateble target used: {}".format(d))

        for d in alldeps:
            if d.need_to_update():
                return True

        return False

    def invoke_function_or_method(self, act):
        if isinstance(act, types.MethodType):
            return act()
        else:
            return act(self)

    def internal_need_if(self):
        return False

    def need_to_update(self):
        if self.need_if is not None:
            return self.need_if(self)
        else:
            return self.internal_need_if()

    def update_if_need(self):
        #has_updated_depends = self.has_updated_depends()
        #need_to_update = self.need_to_update()
        # if self.core.runtime["trace"]:
        #    print(
        #        f"[Trace] {self.tgt} updatable reasons: by_depends:{has_updated_depends}, by_self:{need_to_update}")

        # if has_updated_depends or need_to_update:
        #    return self.invoke_function_or_method(self.update)
        # else:
        #    return True

        if self.recursive_update_needed_request():
            return self.invoke_function_or_method(self.update)
        else:
            return True

    def recurse_update(self, threads=1):
        if "threads" in self.core.runtime:
            threads = self.core.runtime["threads"]

        depset = self.core.depends_as_set(self, incroot=True)
        depset = [self.core.get(d) for d in depset]

        curdep = None
        dtargets = []
        for d in depset:
            def what_to_do(d):
                return d.update_if_need()
            # Это нужно для корректной обработки путей файлов.
            deps_as_targets_names = [self.core.get(d).tgt for d in d.deps]
            dtgt = DependableTarget(name=d.tgt,
                                    deps=deps_as_targets_names,
                                    what_to_do=partial(what_to_do, d),
                                    args=[],
                                    kwargs={})
            dtargets.append(dtgt)
            if d.tgt == self.tgt:
                curdep = dtgt

        success = InverseRecursiveSolver(dtargets, count_of_threads=threads,
                                         trace=self.core.runtime["trace"]).exec()
        if success:
            assert curdep.is_done()
        return success


class Routine(UpdatableTarget):
    __actions__ = {"recurse_update",
                   "recurse_update_get_args", "update", "actlist"}

    def __init__(self,
                 func,
                 deps=[],
                 default_action="update",
                 update_if=lambda s: False,
                 tgt=None,
                 **kwargs):
        if tgt is None:
            tgt = func.__name__
        UpdatableTarget.__init__(self, tgt=tgt, deps=deps,
                                 default_action=default_action, **kwargs)
        self.func = func
        self.update_if = update_if
        self.args = []

    def update(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def self_need(self):
        return self.update_if(self)

    def recurse_update(self, *args, **kwargs):
        self.args = args
        super().recurse_update(**kwargs)


def routine_decorator_do(func=None, deps=[], update_if=lambda s: False, tgt=None, core=default_core()):
    core.add(Routine(func=func, deps=deps, update_if=update_if, tgt=tgt))
    return func


def routine_decorator(func=None, **kwargs):
    if inspect.isfunction(func):
        return routine_decorator_do(func, **kwargs)
    else:
        def decorator(func):
            return routine_decorator_do(func, **kwargs)
        return decorator


# def print_targets_list(target, *args):
#     if core.runtime["debug"]:
#         print("print_targets_list args:{}".format(args))

#     if len(core.targets) == 0:
#         print("targets doesn't founded")
#         return

#     keys = sorted(core.targets.keys())

#     if len(args) > 0:
#         keys = [m for m in keys if re.search(args[0], m)]

#     for k in keys:
#         print(k)


# def print_target_info(taget, *args):
#     if len(args) == 0:
#         licant.error("Need target mnemo")

#     print("name:", core.get(args[0]))
#     print("deps:", sorted(core.get(args[0]).deps))


# def print_deps(taget, *args):
#     if len(args) == 0:
#         name = licant.cli.default_target
#     else:
#         name = args[0]

#     lst = sorted(core.depends_as_set(name))
#     for l in lst:
#         print(l)


# def print_subtree(target, tgt):
#     print(core.subtree(tgt))


# corediag_target = Target(
#     tgt="corediag",
#     deps=[],
#     targets=print_targets_list,
#     tgtinfo=print_target_info,
#     subtree=print_subtree,
#     printdeps=print_deps,
#     actions={"targets", "tgtinfo", "subtree", "printdeps"},
#     __help__="Core state info",
# )

# default_core().add(corediag_target)


def do(target, action=None, args=[], kwargs={}):
    default_core().do(target=target, action=action, args=args, kwargs=kwargs)


def get_target(name):
    return default_core().get(name)


core = default_core()
